@@ -1,7 +1,6 @@
 import click
 
-from sdo.cmd.sood import cmd_ce_vae
-from sdo.cmd.sood import cmd_ae
+from sdo.cmd.sood import cmd_ce_vae, cmd_ae, cmd_threshold
 
 
 @click.group("sood")
@@ -11,3 +10,4 @@ def cli():
 
 cli.add_command(cmd_ce_vae.cli)
 cli.add_command(cmd_ae.cli)
+cli.add_command(cmd_threshold.cli)
