@@ -1,6 +1,6 @@
 import click
 
-from sdo.cmd.sood.ce_vae import cmd_train, cmd_generate, cmd_predict
+from sdo.cmd.sood.ce_vae import cmd_train, cmd_generate, cmd_predict, cmd_reconstruct
 
 
 @click.group("ce_vae")
@@ -11,3 +11,4 @@ def cli():
 cli.add_command(cmd_train.train)
 cli.add_command(cmd_generate.generate)
 cli.add_command(cmd_predict.predict)
+cli.add_command(cmd_reconstruct.reconstruct)
