@@ -1,19 +1,20 @@
 # adjusted from https://github.com/MIC-DKFZ/mood, Credit: D. Zimmerer
 # https://towardsdatascience.com/variational-autoencoder-demystified-with-pytorch-implementation-3a06bee395ed
 import json
-from dateutil.parser import parse
 import datetime
 import os
 import sys
-from math import ceil
 from pathlib import Path
+from typing import Any
 import logging
-
+from dateutil.parser import parse
 import numpy as np
 import pytorch_lightning as pl
 import torch
 import torch.distributions as dist
 import wandb
+from torch.nn import MSELoss, L1Loss
+from torchmetrics import StructuralSimilarityIndexMeasure
 from pytorch_lightning.callbacks import (EarlyStopping, ModelCheckpoint,
                                          ModelSummary)
 from pytorch_lightning.loggers import WandbLogger
@@ -23,13 +24,15 @@ from sdo.sood.data.sdo_ml_v1_dataset import SDOMLv1DataModule
 from sdo.sood.data.sdo_ml_v2_dataset import SDOMLv2DataModule
 from sdo.sood.models.aes import VAE
 from sdo.sood.util.ce_noise import get_square_mask, normalize, smooth_tensor
+from sdo.sood.util.prediction_writer import BatchPredictionWriter
 from sdo.sood.util.utils import (get_smooth_image_gradient, read_config,
-                                 tensor_to_image)
+                                 tensor_to_image, save_image_grid)
 from torch import optim
 from torch.utils.data import DataLoader
 from torchvision.transforms import Compose, Grayscale, Resize, ToTensor
 from torchvision.utils import save_image
-from tqdm import tqdm
+import time
+
 
 logger = logging.getLogger(__name__)
 
@@ -46,7 +49,9 @@ class ceVAE(pl.LightningModule):
         ce_factor,
         score_mode,
         mode,
-        print_every_iter
+        print_every_iter,
+        debug: bool,
+        work_dir
     ):
         super().__init__()
 
@@ -62,6 +67,8 @@ class ceVAE(pl.LightningModule):
         self.lr = lr
         self.vae_loss_ema = 1
         self.theta = 1
+        self.debug = debug
+        self.work_dir = work_dir
 
         self.model = VAE(
             input_size=input_shape[1:], z_dim=z_dim, fmap_sizes=model_feature_map_sizes)
@@ -80,6 +87,7 @@ class ceVAE(pl.LightningModule):
 
     def training_step(self, batch, batch_idx):
         x, attrs = batch  # only inputs no labels
+        t_start = time.process_time()
 
         # VAE Part
         loss_vae = 0
@@ -108,7 +116,8 @@ class ceVAE(pl.LightningModule):
                 ce_tensor = torch.from_numpy(ce_tensor).float()
                 ce_tensor = ce_tensor.type_as(x)
                 inpt_noisy = torch.where(ce_tensor != 0, ce_tensor, x)
-                x_rec_ce, _ = self.model(inpt_noisy)
+                # TODO verify the impact of sampling
+                x_rec_ce, _ = self.model(inpt_noisy, sample=False)
                 rec_loss_ce = self.rec_loss_fn(x_rec_ce, x)
                 loss_ce = rec_loss_ce
 
@@ -151,10 +160,14 @@ class ceVAE(pl.LightningModule):
         self.log('train_loss_vae', loss_vae)
         self.log('train_loss_ce', loss_ce)
 
+        elapsed_time = time.process_time() - t_start
+        self.log('elapsed_time', elapsed_time)
+
         return {"loss": loss, "loss_vae": loss_vae, "loss_ce": loss_ce, "loss_vae_kl": loss_vae_kl, "loss_vae_rec": loss_vae_rec}
 
     def validation_step(self, batch, batch_idx):
         x, _ = batch
+        # TODO why no sampling during validation?
         x_rec, z_dist = self.model(x, sample=False)
         kl_loss = 0
         if self.beta > 0:
@@ -172,104 +185,148 @@ class ceVAE(pl.LightningModule):
                 {"val_images": [wandb.Image(tensor_to_image(
                     x, normalize=False), caption="Input"), wandb.Image(tensor_to_image(
                         x_rec, normalize=False), caption="Output")]})
-        return loss
+        return {"loss": loss, "val_loss_kl": kl_loss, "val_loss_rec": rec_loss}
+
+    def reconstruct(self, batch, sample=False):
+        x, attrs = batch
+        save_path = self.work_dir / Path("reconstructions")
+        if not os.path.exists(save_path):
+            os.makedirs(save_path)
+        logger.info(f"writing reconstructions to {save_path}")
+
+        ssim = StructuralSimilarityIndexMeasure()
+        mse = MSELoss()
+        mae = L1Loss()
+        loss_path = save_path / Path("reconstruction_loss.csv")
+        if not os.path.exists(loss_path):
+            with open(loss_path, "a") as target_file:
+                target_file.write(
+                    f"pixel_pred_path,ssim,mse,mae,t_obs,wavelength\n")
+        with torch.no_grad():
+            mu, std = self.model.encode(x)
+            z_dist = dist.Normal(mu, std)
+            if sample:
+                z_sample = z_dist.rsample()
+            else:
+                z_sample = mu
+            x_rec = self.model.decode(z_sample)
+
+            for idx in range(len(x_rec)):
+                x_i, x_rec_i, mu_i, std_i = x[idx], x_rec[idx], mu[idx], std[idx]
+
+                t_obs = attrs["T_OBS"][idx]
+                wavelength = attrs["WAVELNTH"][idx]
+                timestamp = parse(t_obs)
+                file_name = f"{timestamp.strftime(folder_time_format)}_{wavelength}A"
+                input_img = tensor_to_image(x[idx])
+                input_img.save(save_path / Path(
+                    f"{file_name}_src.png"))
+                rec_img = tensor_to_image(x_rec_i)
+                rec_img.save(save_path / Path(
+                    f"{file_name}_rec.png"))
+
+                torch.save(mu_i, save_path / Path(
+                    f"{file_name}_mu.pt"))
+                torch.save(std_i, save_path / Path(
+                    f"{file_name}_std.pt"))
+
+                x_rec_i = x_rec_i[None, :, :, :]
+                x_i = x_i[None, :, :, :]
+                x_ssim = ssim(x_rec_i, x_i)
+                x_mse = mse(x_rec_i, x_i)
+                x_mae = mae(x_rec_i, x_i)
+                with open(loss_path, "a") as target_file:
+                    target_file.write(
+                        f"{file_name},{str(x_ssim.numpy())},{str(x_mse.numpy())},{str(x_mae.numpy())},{t_obs},{wavelength}\n")
 
     def test_step(self, batch, batch_idx):
         logger.warn("test step not defined")
         pass
 
-    def forward(self, x):
+    def predict_step(self, batch: Any, batch_idx: int, dataloader_idx: int = 0) -> Any:
+        return self(batch, batch_idx=batch_idx)
+
+    def forward(self, batch, batch_idx: int = None):
         if self.mode == "sample":
-            return self.score_sample(x)
+            return self.score_sample(batch)
         elif self.mode == "pixel":
-            return self.score_pixels(x)
+            return self.score_pixels(batch, batch_idx=batch_idx)
         else:
             raise ValueError(f"invalid mode {self.mode}")
 
-    def generate(self, n_samples=16, mu=None, std=None):
-        if mu is None:
-            mu = torch.zeros_like(torch.empty(self.z_dim, 1, 1))
-        if std is None:
-            std = torch.ones_like(torch.empty(self.z_dim, 1, 1))
+    def generate(self, n_samples=1, n_iter=128, mu=None, std=None, with_cm=False):
+        output_dir = Path(self.work_dir) / Path("generated")
+        if not os.path.exists(output_dir):
+            os.makedirs(output_dir)
+        logger.info(f"writing generated examples to {output_dir}")
+        for i in range(n_iter):
+            if mu is None:
+                mu = torch.zeros_like(torch.empty(self.z_dim, 1, 1))
+            if std is None:
+                std = torch.ones_like(torch.empty(self.z_dim, 1, 1))
 
-        p = torch.distributions.Normal(mu, std)
-        z = p.rsample((n_samples,))
-
-        with torch.no_grad():
-            pred = self.model.decode(z.to(self.device)).cpu()
-
-        file_name = Path(self.work_dir) / Path("generated") / \
-            (datetime.datetime.now().isoformat() + "_generated.jpeg")
-        save_image(pred, file_name, normalize=True)
-
-    def score_sample(self, data):
-        orig_shape = data.shape
-        to_transforms = torch.nn.Upsample(
-            (self.input_shape[2], self.input_shape[3]), mode="bilinear")
-
-        data_tensor = data
-        data_tensor = to_transforms(data_tensor[None])[0]
-        slice_scores = []
-
-        for i in range(ceil(orig_shape[0] / self.batch_size)):
-            batch = data_tensor[i *
-                                self.batch_size: (i + 1) * self.batch_size].unsqueeze(1)
-            # batch = batch * 2 - 1
+            p = torch.distributions.Normal(mu, std)
+            z = p.rsample((n_samples,))
 
             with torch.no_grad():
-                inpt = batch.to(self.device).float()
-                x_rec, z_dist = self.model(inpt, sample=False)
-                kl_loss = self.kl_loss_fn(z_dist, sum_samples=False)
-                rec_loss = self.rec_loss_fn(x_rec, inpt, sum_samples=False)
-                img_scores = kl_loss * self.beta + rec_loss * self.theta
+                pred = self.model.decode(z.to(self.device)).cpu()
 
-            slice_scores += img_scores.cpu().tolist()
+            file_name = output_dir / \
+                f"{datetime.datetime.now().strftime(folder_time_format)}_{i}_generated.jpeg"
 
-        return np.max(slice_scores)
+            if with_cm:
+                from PIL import Image
+                from sunpy.visualization.colormaps import cm
+                from torchvision.utils import make_grid
 
-    def score_pixels(self, data):
-        orig_shape = data.shape
-        to_transforms = torch.nn.Upsample(
-            (self.input_shape[2], self.input_shape[3]), mode="bilinear")
-        from_transforms = torch.nn.Upsample(
-            (orig_shape[1], orig_shape[2]), mode="bilinear")
-        data_tensor = data
-        data_tensor = to_transforms(data_tensor[None])[0]
-        target_tensor = torch.zeros_like(data_tensor)
+                grid = make_grid(pred, normalize=True)
+                ndarr = grid.mul(255).add_(0.5).clamp_(0, 255).permute(
+                    1, 2, 0).to("cpu", torch.uint8).numpy()
+                m = cm.cmlist.get('sdoaia%d' % int(171))
+                v = np.squeeze(ndarr[:, :, 0])
+                v = m(v)
+                v = (v[:, :, :3]*255).astype(np.uint8)
+                im = Image.fromarray(v)
+                im.save(file_name)
+            else:
+                save_image(pred, file_name, normalize=True, scale_each=True)
 
-        for i in range(ceil(orig_shape[0] / self.batch_size)):
-            batch = data_tensor[i *
-                                self.batch_size: (i + 1) * self.batch_size].unsqueeze(1)
+    def score_sample(self, batch):
+        data, _ = batch
+        x_rec, z_dist = self.model(data, sample=False)
+        kl_loss = self.kl_loss_fn(z_dist, sum_samples=False)
+        rec_loss = self.rec_loss_fn(x_rec, data, sum_samples=False)
+        sample_scores = kl_loss * self.beta + rec_loss * self.theta
 
-            inpt = batch.to(self.device).float()
-            x_rec, z_dist = self.model(inpt, sample=False)
+        return sample_scores.cpu().tolist()
 
+    def score_pixels(self, batch, batch_idx: int = None):
+        data, _ = batch
+        with torch.enable_grad():
+            x_rec, _ = self.model(data, sample=False)
             if self.score_mode == "combi":
-                rec = torch.pow((x_rec - inpt), 2).detach().cpu()
+                rec = torch.pow((x_rec - data), 2).detach().cpu()
                 rec = torch.mean(rec, dim=1, keepdim=True)
 
                 def __err_fn(x):
                     x_r, z_d = self.model(x, sample=False)
                     loss = self.kl_loss_fn(z_d)
                     return loss
-
                 loss_grad_kl = (
                     get_smooth_image_gradient(
-                        model=self.model, inpt=inpt, err_fn=__err_fn, grad_type="vanilla", n_runs=2
+                        model=self.model, inpt=data, err_fn=__err_fn, grad_type="vanilla", n_runs=2
                     )
                     .detach()
                     .cpu()
                 )
-                loss_grad_kl = torch.mean(loss_grad_kl, dim=1, keepdim=True)
-
+                loss_grad_kl = torch.mean(
+                    loss_grad_kl, dim=1, keepdim=True)
                 pixel_scores = smooth_tensor(
                     normalize(loss_grad_kl), kernel_size=8) * rec
-
             elif self.score_mode == "rec":
-                rec = torch.pow((x_rec - inpt), 2).detach().cpu()
+                rec = torch.pow((x_rec - data), 2).detach().cpu()
                 rec = torch.mean(rec, dim=1, keepdim=True)
                 pixel_scores = rec
-
             elif self.score_mode == "grad":
                 def __err_fn(x):
                     x_r, z_d = self.model(x, sample=False)
@@ -277,34 +334,29 @@ class ceVAE(pl.LightningModule):
                     rec_loss_ = self.rec_loss_fn(x_r, x)
                     loss_ = kl_loss_ * self.beta + rec_loss_ * self.theta
                     return torch.mean(loss_)
-
                 loss_grad_kl = (
                     get_smooth_image_gradient(
-                        model=self.model, inpt=inpt, err_fn=__err_fn, grad_type="vanilla", n_runs=2
+                        model=self.model, inpt=data, err_fn=__err_fn, grad_type="vanilla", n_runs=2
                     )
                     .detach()
                     .cpu()
                 )
-                loss_grad_kl = torch.mean(loss_grad_kl, dim=1, keepdim=True)
-
+                loss_grad_kl = torch.mean(
+                    loss_grad_kl, dim=1, keepdim=True)
                 pixel_scores = smooth_tensor(
                     normalize(loss_grad_kl), kernel_size=8)
+            pixel_scores = pixel_scores.detach().cpu()
+            if self.debug:
+                logger.info(
+                    f"Writing debug information about predictions to {self.work_dir}")
+                save_image_grid(data, name="Input", save_dir=self.work_dir, image_args={
+                    "normalize": True}, n_iter=batch_idx)
+                save_image_grid(x_rec, name="Reconstruction", save_dir=self.work_dir, image_args={
+                    "normalize": True}, n_iter=batch_idx)
+                save_image_grid(pixel_scores, name="Scores", save_dir=self.work_dir, image_args={
+                    "normalize": True}, n_iter=batch_idx)
 
-            # save_image_grid(inpt, name="Input", save_dir=self.work_dir, image_args={
-            #     "normalize": True}, n_iter=index)
-            # save_image_grid(x_rec, name="Output", save_dir=self.work_dir, image_args={
-            #     "normalize": True}, n_iter=index)
-            # save_image_grid(pixel_scores, name="Scores", save_dir=self.work_dir, image_args={
-            #     "normalize": True}, n_iter=index)
-
-            target_tensor[i * self.batch_size: (
-                i + 1) * self.batch_size] = pixel_scores.detach().cpu()[:, 0, :]
-
-        target_tensor = from_transforms(target_tensor[None])[0]
-        # TODO rather normalize over the whole dataset rather than a single image
-        # save_image(target_tensor, file_name, normalize=True)
-
-        return target_tensor
+        return pixel_scores
 
     @staticmethod
     def kl_loss_fn(z_post, sum_samples=True, correct=False):
@@ -333,21 +385,6 @@ class ceVAE(pl.LightningModule):
         else:
             return -log_p_x_z
 
-    @staticmethod
-    def get_inpt_grad(model, inpt, err_fn):
-        model.zero_grad()
-        inpt = inpt.detach()
-        inpt.requires_grad = True
-
-        err = err_fn(inpt)
-        err.backward()
-
-        grad = inpt.grad.detach()
-
-        model.zero_grad()
-
-        return torch.abs(grad.detach())
-
     @staticmethod
     def geco_beta_update(beta, error_ema, goal, step_size, min_clamp=1e-10, max_clamp=1e4, speedup=None):
         constraint = (error_ema - goal).detach()
@@ -361,24 +398,22 @@ class ceVAE(pl.LightningModule):
             beta = np.min((beta.item(), max_clamp))
         return beta
 
-    @staticmethod
-    def get_ema(new, old, alpha):
-        if old is None:
-            return new
-        return (1.0 - alpha) * new + alpha * old
-
 
 folder_time_format = "%Y%m%d-%H%M%S"
 
 
 def main(
     run: str = "train",
-    config_file: Path = Path("./config/defaults.yaml")
+    config_file: Path = Path("./config/defaults.yaml"),
+    predict_mode: str = None,
+    config_overrides: dict = None
 ):
-    config = read_config(config_file)
+    config = read_config(config_file, config_overrides)
     logger.info("found config")
     logger.info(json.dumps(config, indent=2))
 
+    predict_mode = predict_mode or config.predict.mode.value
+
     current_run_name = f"{datetime.datetime.now().strftime(folder_time_format)}_cevae"
     work_dir = Path(config.log_dir.value) / Path(current_run_name)
     if not os.path.exists(work_dir):
@@ -389,7 +424,7 @@ def main(
 
     if config.model.load_path.value is not None:
         cevae_algo = ceVAE.load_from_checkpoint(
-            config.model.load_path.value, mode=config.predict.mode.value)
+            config.model.load_path.value, mode=predict_mode,  debug=config.debug.value, work_dir=work_dir)
     else:
         cevae_algo = ceVAE(
             input_shape,
@@ -400,8 +435,10 @@ def main(
             beta=config.train.beta.value,
             ce_factor=config.model.ce_factor.value,
             score_mode=config.predict.score_mode.value,
-            mode=config.predict.mode.value,
-            print_every_iter=config.train.print_every_iter.value
+            mode=predict_mode,
+            print_every_iter=config.train.print_every_iter.value,
+            debug=config.debug.value,
+            work_dir=work_dir
         )
 
     data_module = None
@@ -456,21 +493,25 @@ def main(
                                                 train_val_split_temporal_chunk_size=config.data.sdo_ml_v2.train_val_split_temporal_chunk_size.value,
                                                 sampling_strategy=config.data.sdo_ml_v2.sampling_strategy.value,
                                                 mask_limb=config.data.sdo_ml_v2.mask_limb.value,
-                                                mask_limb_radius_scale_factor=config.data.sdo_ml_v2.mask_limb_radius_scale_factor.value)
+                                                mask_limb_radius_scale_factor=config.data.sdo_ml_v2.mask_limb_radius_scale_factor.value,
+                                                reduce_memory=config.data.sdo_ml_v2.reduce_memory)
 
         # config_exclude_keys are already logged from config directly
         wandb_logger = WandbLogger(
-            project="sdo-sood", log_model="all", config_exclude_keys=["input_shape",
-                                                                      "lr",
-                                                                      "z_dim",
-                                                                      "model_feature_map_sizes",
-                                                                      "use_geco",
-                                                                      "beta",
-                                                                      "ce_factor",
-                                                                      "score_mode",
-                                                                      "mode",
-                                                                      "print_every_iter"])
-        wandb_logger.experiment.config.update(config)
+            project="sdo-sood",
+            log_model="all",
+            id=config.train.wandb_run_id.value,
+            config_exclude_keys=["input_shape",
+                                 "lr",
+                                 "z_dim",
+                                 "model_feature_map_sizes",
+                                 "use_geco",
+                                 "beta",
+                                 "ce_factor",
+                                 "score_mode",
+                                 "mode",
+                                 "print_every_iter"])
+        wandb_logger.experiment.config.update(config, allow_val_change=True)
 
         profiler = None
         if config.train.profile.value == True:
@@ -509,7 +550,7 @@ def main(
         cevae_algo.eval()
         cevae_algo.generate()
 
-    if run == "predict":
+    if run == "predict" or run == "reconstruct":
         cevae_algo.eval()
         pred_dir = config.predict.pred_dir.value
         if pred_dir is None:
@@ -548,8 +589,9 @@ def main(
                                                 storage_driver=config.data.sdo_ml_v2.storage_driver.value,
                                                 num_workers=config.data.num_data_loader_workers.value,
                                                 pin_memory=False,
+                                                obs_times=config.data.sdo_ml_v2.obs_times.value,
                                                 target_size=input_shape[2],
-                                                batch_size=1,
+                                                batch_size=config.data.batch_size.value,
                                                 prefetch_factor=config.data.prefetch_factor.value,
                                                 channel=config.data.sdo_ml_v2.channel.value,
                                                 freq=config.data.sdo_ml_v2.freq.value,
@@ -560,29 +602,14 @@ def main(
                                                 test_end=config.data.sdo_ml_v2.test_end_date.value,
                                                 skip_train_val=True,
                                                 mask_limb=config.data.sdo_ml_v2.mask_limb.value,
-                                                mask_limb_radius_scale_factor=config.data.sdo_ml_v2.mask_limb_radius_scale_factor.value)
+                                                mask_limb_radius_scale_factor=config.data.sdo_ml_v2.mask_limb_radius_scale_factor.value,
+                                                reduce_memory=config.data.sdo_ml_v2.reduce_memory)
                 data_loader = data_module.predict_dataloader()
-        for index, batch in tqdm(enumerate(data_loader)):
-            # TODO attrs are not available for ImageParameterDataset
-            img, attrs = batch
-            timestamp = parse(attrs["T_OBS"][0])
-            wavelength = attrs["WAVELNTH"][0]
-            file_name = Path(
-                f"{timestamp.strftime(folder_time_format)}_{wavelength}A.png")
-            file_path = Path(pred_dir) / file_name
-            if config.predict.mode.value == "pixel":
-                pixel_scores = cevae_algo.score_pixels(img[0])
-                # TODO rather normalize over the full dataset
-                save_src_image = False  # TODO make configurable
-                if save_src_image:
-                    src_file_name = Path(
-                        f"{timestamp.strftime(folder_time_format)}_{wavelength}A_src.png")
-                    src_file_path = Path(pred_dir) / src_file_name
-                    save_image(img[0], src_file_path)
-                save_image(pixel_scores, file_path, normalize=True)
-            if config.predict.mode.value == "sample":
-                # TODO make it work for batches instead of single images
-                sample_score = cevae_algo.score_sample(img[0])
-                with open(os.path.join(pred_dir, "predictions.txt"), "a") as target_file:
-                    target_file.write(file_path.name + "," +
-                                      str(sample_score) + "\n")
+        if run == "reconstruct":
+            for _, batch in enumerate(data_module.predict_dataloader()):
+                cevae_algo.reconstruct(batch)
+        elif run == "predict":
+            logger.info(f"logging predictions to {pred_dir}")
+            trainer = pl.Trainer(
+                gpus=config.devices.gpus.value, accelerator="auto", callbacks=[BatchPredictionWriter(output_dir=pred_dir, mode=predict_mode, save_src_img=config.predict.save_src_img.value)])
+            trainer.predict(cevae_algo, data_loader, return_predictions=False)
