@@ -59,7 +59,9 @@ class SDOMLv2NumpyDataset(Dataset):
             irradiance_channel: str,
             goes_cache_dir: str,
             transforms: list,
-            cache_max_size: int):
+            cache_max_size: int,
+            reduce_memory: bool,
+            obs_times: list):
         """Dataset which loads the SDO Ml v2 dataset from a zarr directory.
         """
 
@@ -105,6 +107,12 @@ class SDOMLv2NumpyDataset(Dataset):
             attrs = dict(arr.attrs)
 
             images, attrs = self.data_quality_check(channel, images, attrs)
+            if reduce_memory:
+                minimal_keys = ["T_OBS", "WAVELNTH"]
+                logger.info(
+                    f"reducing memory by retaining only the following attributes {minimal_keys}")
+                attrs = {key: attrs[key] for key in minimal_keys}
+
             if freq:
                 images, attrs = self.temporal_downsampling(
                     channel, start, end, freq, images, attrs)
@@ -115,6 +123,10 @@ class SDOMLv2NumpyDataset(Dataset):
             if irradiance is not None:
                 images, attrs = self.irradiance_filtering(
                     channel, irradiance, irradiance_channel, goes_cache_dir, images, attrs)
+
+            if obs_times is not None:
+                images, attrs = self.t_obs_filtering(
+                    channel, obs_times, images, attrs)
             all_images.append(images)
             all_attrs.append(attrs)
 
@@ -177,6 +189,20 @@ class SDOMLv2NumpyDataset(Dataset):
 
         return images, attrs
 
+    def t_obs_filtering(self, channel, obs_times, images, attrs):
+        logger.info(
+            f"filtering for {len(obs_times)} obs times")
+
+        obs_times = pd.to_datetime(obs_times, format=None, utc=True)
+        df_time = self.obs_time_as_df(channel, attrs)
+        filter = np.isin(df_time['Time'], obs_times)
+
+        time_index = df_time['Time'].index[filter].tolist()
+        images = images[time_index, :, :]
+        attrs = {keys: [values[idx] for idx in time_index]
+                 for keys, values in attrs.items()}
+        return images, attrs
+
     def temporal_filtering(self, channel, start, end, images, attrs):
         logger.info(
             f"filtering data between {start} and {end}")
@@ -282,6 +308,9 @@ class SDOMLv2NumpyDataset(Dataset):
 
 # TODO are these values still applicable for the new correction factors?
 # Same preprocess as github.com/i4Ds/SDOBenchmark
+# Notably in SDOBenchmark some of the preprocessing seems to be different from the preprocessing in Helioviewer
+# https://github.com/Helioviewer-Project/jp2gen/blob/master/idl/sdo/aia/hvs_default_aia.pro
+# e.g. AIA 171 has log10 scaling in SDO benchmark and sqrt scaling in Helioviewer
 CHANNEL_PREPROCESS = {
     "94A": {"min": 0.1, "max": 800, "scaling": "log10"},
     "131A": {"min": 0.7, "max": 1900, "scaling": "log10"},
@@ -330,7 +359,8 @@ def get_default_transforms(target_size=256, channel="171", mask_limb=False, radi
     preprocess_config = CHANNEL_PREPROCESS[channel]
 
     if preprocess_config["scaling"] == "log10":
-        # TODO does it make sense to use vflip(x) in order to align the solar North as in JHelioviewer? otherwise this has to be done during inference
+        # TODO does it make sense to use vflip(x) in order to align the solar North as in JHelioviewer?
+        # otherwise this has to be done during inference
         def lambda_transform(x): return torch.log10(torch.clamp(
             x,
             min=preprocess_config["min"],
@@ -339,6 +369,15 @@ def get_default_transforms(target_size=256, channel="171", mask_limb=False, radi
         mean = math.log10(preprocess_config["min"])
         std = math.log10(preprocess_config["max"]) - \
             math.log10(preprocess_config["min"])
+    elif preprocess_config["scaling"] == "sqrt":
+        def lambda_transform(x): return torch.sqrt(torch.clamp(
+            x,
+            min=preprocess_config["min"],
+            max=preprocess_config["max"],
+        ))
+        mean = math.sqrt(preprocess_config["min"])
+        std = math.sqrt(preprocess_config["max"]) - \
+            math.sqrt(preprocess_config["min"])
     else:
         def lambda_transform(x): return torch.clamp(
             x,
@@ -413,6 +452,7 @@ class SDOMLv2DataModule(pl.LightningDataModule):
                  test_start=None,
                  test_end=None,
                  freq=None,
+                 obs_times: list = None,
                  irradiance: float = None,
                  irradiance_channel: str = "xrsb",
                  goes_cache_dir: str = None,
@@ -421,9 +461,11 @@ class SDOMLv2DataModule(pl.LightningDataModule):
                  train_val_split_ratio: float = 0.7,
                  train_val_split_temporal_chunk_size: str = "14d",
                  skip_train_val: bool = False,
+                 skip_test: bool = False,
                  sampling_strategy: str = "chunk",
                  mask_limb: bool = False,
-                 mask_limb_radius_scale_factor: float = 1.0
+                 mask_limb_radius_scale_factor: float = 1.0,
+                 reduce_memory: bool = False
                  ):
         """
         Creates a LightningDataModule for the SDO Ml v2 dataset
@@ -451,6 +493,7 @@ class SDOMLv2DataModule(pl.LightningDataModule):
             test_start (str, optional): [Allows to restrict the dataset temporally]. Defaults to None.
             test_end (str, optional): [Allows to restrict the dataset temporally]. Defaults to None.
             freq (str, optional): [Allows to downsample the dataset temporally, should be bigger than the min interval for the observed channel. When using freq, start and end should also be specified for train and test]. Defaults to None.
+            obs_times (list, optional): [Allows to filter for specific observation times]. Defaults to None.
             irradiance (float, optional): Allows to filter by irradiance (all images with smaller or equal values will be in the resulting dataset). Defaults to None.
             irradiance_channel (str, optional): Either xrsa or xrsb. Refer to the GOES documentation. Defaults to "xrsb".
             goes_cache_dir (str, optional): Path to the cached GOES values (downloaded with sdo-cli goes download --help). Defaults to None.
@@ -459,9 +502,11 @@ class SDOMLv2DataModule(pl.LightningDataModule):
             train_val_split_ratio (float, optional): [Split-ratio for the train-validation split]. Defaults to 0.7.
             train_val_split_temporal_chunk_size (str, optional): [Temporal chunks for the train-validation splits]. Defaults to "14d".
             skip_train_val (bool, optional): [Whether to skip loading the train and validation sets]: Defaults to False.
+            skip_test (bool, optional): [Whether to skip loading the test set]: Defaults to False.
             sampling_strategy (str, optional): [Which sampling strategy to use (chunk or default)]: Defaults to "chunk".
             mask_limb (bool, optional): [Whether to mask the solar limb]. Defaults to False.
             mask_limb_radius_scale_factor (float, optional): [Allows to scale the radius that is used for masking the solar limb]. Defaults to 1.0.
+            reduce_memory (bool, optional): [Reduces memory by only retaining required attributes]. Defaults to False.
         """
 
         super().__init__()
@@ -469,6 +514,26 @@ class SDOMLv2DataModule(pl.LightningDataModule):
         transforms = get_default_transforms(
             target_size=target_size, channel=channel, mask_limb=mask_limb, radius_scale_factor=mask_limb_radius_scale_factor)
 
+        # TODO this can be achieved with setup()
+        # https://pytorch-lightning.readthedocs.io/en/latest/api/pytorch_lightning.core.hooks.DataHooks.html#pytorch_lightning.core.hooks.DataHooks.setup
+        if not skip_test:
+            self.dataset_test = SDOMLv2NumpyDataset(
+                storage_root=storage_root,
+                storage_driver=storage_driver,
+                cache_max_size=cache_max_size,
+                year=test_year,
+                start=test_start,
+                end=test_end,
+                freq=freq,
+                obs_times=obs_times,
+                irradiance=irradiance,
+                irradiance_channel=irradiance_channel,
+                goes_cache_dir=goes_cache_dir,
+                channel=channel,
+                transforms=transforms,
+                reduce_memory=reduce_memory
+            )
+
         if not skip_train_val:
             dataset = SDOMLv2NumpyDataset(
                 storage_root=storage_root,
@@ -478,29 +543,14 @@ class SDOMLv2DataModule(pl.LightningDataModule):
                 start=train_start,
                 end=train_end,
                 freq=freq,
+                obs_times=obs_times,
                 irradiance=irradiance,
                 irradiance_channel=irradiance_channel,
                 goes_cache_dir=goes_cache_dir,
                 channel=channel,
                 transforms=transforms,
+                reduce_memory=reduce_memory
             )
-
-        self.dataset_test = SDOMLv2NumpyDataset(
-            storage_root=storage_root,
-            storage_driver=storage_driver,
-            cache_max_size=cache_max_size,
-            year=test_year,
-            start=test_start,
-            end=test_end,
-            freq=freq,
-            irradiance=irradiance,
-            irradiance_channel=irradiance_channel,
-            goes_cache_dir=goes_cache_dir,
-            channel=channel,
-            transforms=transforms,
-        )
-
-        if not skip_train_val:
             if train_val_split_strategy == "random":
                 num_samples = len(dataset)
                 splits = [int(math.floor(num_samples*train_val_split_ratio)),
@@ -538,7 +588,7 @@ class SDOMLv2DataModule(pl.LightningDataModule):
             sampler = SequenceInChunkSampler(self.dataset_val)
         return DataLoader(self.dataset_val, batch_size=self.batch_size,
                           shuffle=self.shuffle,
-                          num_workers=self.num_workers,
+                          num_workers=math.ceil(self.num_workers/2),
                           pin_memory=self.pin_memory,
                           drop_last=self.drop_last,
                           prefetch_factor=self.prefetch_factor,
