@@ -1,3 +1,5 @@
+from matplotlib.colors import LinearSegmentedColormap
+from sunpy.visualization.colormaps import cm
 from munch import DefaultMunch
 import json
 import random
@@ -223,7 +225,7 @@ def save_image_grid(tensor, name, save_dir, n_iter=None, prefix=False, iter_form
         tensor = tensor[:, 0:1, ]
 
     if n_iter is not None:
-        name = name_and_iter_to_filename(name=name, n_iter=n_iter, ending=".png", iter_format=iter_format,
+        name = name_and_iter_to_filename(name=name, n_iter=n_iter, extension=".png", iter_format=iter_format,
                                          prefix=prefix)
     elif not name.endswith(".png"):
         name += ".png"
@@ -239,12 +241,12 @@ def save_image_grid(tensor, name, save_dir, n_iter=None, prefix=False, iter_form
     return str(img_file)
 
 
-def name_and_iter_to_filename(name, n_iter, ending, iter_format="{:05d}", prefix=False):
+def name_and_iter_to_filename(name, n_iter, extension, iter_format="{:05d}", prefix=False):
     iter_str = iter_format.format(n_iter)
     if prefix:
-        name = iter_str + "_" + name + ending
+        name = iter_str + "_" + name + extension
     else:
-        name = name + "_" + iter_str + ending
+        name = name + "_" + iter_str + extension
 
     return name
 
@@ -286,7 +288,7 @@ def merge_config(user: dict, default: dict) -> dict:
     return user
 
 
-def read_config(config_file: Path) -> dict:
+def read_config(config_file: Path, overrides: dict = None) -> dict:
     config_file = Path(os.path.expanduser(config_file))
     with open(config_file, "r") as f:
         user_config = yaml.safe_load(f)
@@ -297,5 +299,28 @@ def read_config(config_file: Path) -> dict:
             default_config = yaml.safe_load(f)
             user_config = merge_config(user_config, default_config)
 
+    if overrides:
+        user_config = merge_config(overrides, user_config)
+
     # Hack to convert dict to object
     return DefaultMunch.fromDict(user_config)
+
+
+# Channels that correspond to HMI Magnetograms
+HMI_WL = ['Bx', 'By', 'Bz']
+
+# A colormap for visualizing HMI
+HMI_CM = LinearSegmentedColormap.from_list(
+    "bwrblack", ["#0000ff", "#000000", "#ff0000"])
+
+
+def channel_to_map(name):
+    """Given channel name, return colormap"""
+    return HMI_CM if name in HMI_WL else cm.cmlist.get('sdoaia%d' % int(name))
+
+
+def colorize_sdo(X, channel=171):
+    """Given image and channel, visualize results"""
+    cm = channel_to_map(channel)
+    Xcv = cm(X)
+    return (Xcv[:, :, :3]*255).astype(np.uint8)
